//: Playground - noun: a place where people can play

import Cocoa

/*
 - 8Queenパズルとは、
 - チェスのクイーンは、チェス盤上で最も自由に動けまわれるコマです。
 - 縦横斜めの8方向に、任意の数だけ駒を進めることが出来ます。将棋でいう飛車と角を合わせた動きです。
 - このクイーンを8x8のチェス盤上に8個、お互いに取られないように配置するのが8Queenパズルです。
 
 - 8Queenパズルの解法は、しらみ潰し的な試行錯誤を繰り返さないと解けないそうです。
 -- このため、しらみつぶしを効率よく行う、アルゴリズム、バックトラッキングのサンプルコードとしてよく使われています。
 */


/*
 作る部品。
 
 - queenを置くチェス盤 board
 -- queenを置くごとに、新しいチェス盤を生成する。
 -- structで定義してletで使うようにする。これは、関数プログラミング的にするためでなく、後でバックトラックする時やViewを作る時のため。
 -- boardはのそれぞれの升は、「からで配置可能、からで配置不可能、すでにQueenが配置されている」の3つの状態をもつ。
 -- boardは全体で、お互いに取られない状態であるか、取られる状態のqeenが配置されているか、
 -- 配置されているqueenの数、からで配置可能なマスの数、からで配置不可能なマスの数を返す。<--マスの集合を返す。
 -- 盤面はn*nとする。これはデバッグをしやすくするため。
 
 - 最初に盤を作り、1個のqueenを置く、
 - 次に配置可能な場所のリストを次々と試す、
 -- 
 
 */

// queenを置くチェス盤。
struct EQBoard
{
    let sizeN : Int // ボードの一辺の大きさ
    let queens  : [Int]
    
    
    // 初期化
    init(size:Int)
    {
        self.init(size:size, queens:[])
    }

    init(size:Int, queens: [Int])
    {
        precondition(queens.count <= size)        
        precondition(queens.filter{$0 < 0}.count == 0)
        precondition(queens.filter{$0 >= size}.count == 0)
        
        self.sizeN = size
        self.queens = queens
    }

    func resolve() -> [EQBoard]
    {
        precondition(queens.count <= self.sizeN)        
                
        func resolve_intarnal( _ count:Int, _ boards:[EQBoard]) -> [EQBoard]
        {
            if count == 0
            {
                return boards;
            }
            
            // 与えられた板の次のrawの配置の配列を計算する
            return resolve_intarnal( count - 1, 
                                     boards.reduce([])// 計算結果を配列にひとまとめにする
                                     {
                                        let i = $1
                                        
                                        return $0 + ( 0..<i.sizeN ) // 横のマスの数だけ、置いて試す
                                            .flatMap
                                            {                        
                                                i.tryAddAQueen($0) ? i.boardAddAQueen($0) : nil
                                            }
                                    })
        }
        
        return resolve_intarnal(self.sizeN, [self])
    }
    
    
    
    // queenを1個追加した盤を返す
    func boardAddAQueen(_ newQueen: Int) -> EQBoard
    {
        precondition(newQueen >= 0)        
        precondition(newQueen < self.sizeN)        
        precondition(queens.count < self.sizeN)        
        
        return EQBoard(size:self.sizeN, queens: self.queens + [newQueen])
    }
    
    
    // 配置可能かのチェックを行う
    func tryAddAQueen(_ newQueen: Int) -> Bool
    {
        precondition(newQueen >= 0)        
        precondition(newQueen < self.sizeN)        

        // 縦方向と左右の斜めにQueenがないかチェックする
        return ( self.queens.enumerated().map
                    { 
                        ( $0.element == newQueen)                                  || /* 縦が一致       */
                        ( $0.element == newQueen - self.queens.count + $0.offset ) || /* 斜め左方向が一致 */
                        ( $0.element == newQueen + self.queens.count - $0.offset ) ?  /* 斜め右方向が一致 */
                         false : true
                    }.filter{ $0==false }.count ) == 0        
    }
    
    // debug用出力
    func debugPrint()
    {
        // 数列に対するmap&reduceを行い、デバッグ文字列を作成する
        print(
            ( 0..<queens.count ).map
                {
                    // ここで引数に名前をつけないと内側のクロージャでアクセスできない。
                    // $$.$0のような表記で外側の環境へアクセスする方法はないのだろうか？それができれば、"return"文も省略できる。
                    let i = $0 
                    return ( (0..<sizeN).map{ $0 == queens[i] ? " q" : " ." }.reduce(""){ $0 + $1 } ) + "\n"
                
                }.reduce(""){ $0 + $1 }
        )
    }
}

let queen = EQBoard(size:8)
let QueenAnser = queen.resolve()

QueenAnser.map{ $0.debugPrint() }
print("---\(QueenAnser.count)")

